#include "pqtree.h"

pqtree::pqtree(uint32_t cnt) :
	m_leaves(cnt), m_nodes(cnt + 3), m_freenode(0)
{
	for (uint32_t i = 0; i < m_nodes.size(); i++) {
		release_node(&m_nodes[i]);
	}
	m_sentinel = new_node(pqtype_p);
	pqnode *root = new_node(pqtype_p);

	root->insert_after(m_sentinel, 0);
	
	for (uint32_t i = 0; i < cnt; i++) {
		pqnode *n = &m_leaves[i];
		n->id = i;
		n->type = pqtype_leaf;
		n->insert_after(root, root->last);
	}
}

pqtree::pqtree(const pqtree & t) :
	m_leaves(t.m_leaves.size()), m_nodes(t.m_leaves.size() + 3), m_freenode(0)
{
	for (uint32_t i = 0; i < m_nodes.size(); i++) {
		release_node(&m_nodes[i]);
	}
	m_sentinel = new_node(pqtype_p);

	for (uint32_t i = 0; i < m_leaves.size(); i++) {
		pqnode *n = &m_leaves[i];
		n->id = i;
		n->type = pqtype_leaf;
	}

	pqnode *cur = t.m_sentinel;
	pqnode *child = cur->first;

	pqnode *clonenode = m_sentinel;
	pqnode *clonechild;

	while (cur) {
		if (child->type == pqtype_leaf)
			clonechild = &m_leaves[child->id];
		else
			clonechild = new_node(child->type);
		clonechild->insert_after(clonenode, clonenode->last);

		if (child->first) {
			cur = child;
			child = child->first;
			clonenode = clonechild;
		}
		else {
			child = child->right;
			while (!child && cur) {
				child = cur->right;
				cur = cur->parent;
				clonenode = clonenode->parent;
			}
		}
	}
}

void
pqtree::reset_counter()
{
	m_queue.reset();
	for (uint32_t i = 0; i < m_leaves.size(); i++) {
		m_queue.push(&m_leaves[i]);
	}

	while (!m_queue.empty()) {
		pqnode *n = m_queue.pop();
		n->visited_count = 0;
		m_queue.push(n->parent);
	}
}

pqnode *
pqtree::find_root(uintvector & order)
{
	reset_counter();

	m_queue.reset();


	for (uint32_t i = 0; i < m_leaves.size(); i++) {
		m_leaves[i].pertinent_leaf_count = 0;
		m_queue.push(&m_leaves[i]);
	}

	for (uint32_t i = 0; i < order.size(); i++) {
		m_leaves[order[i]].pertinent_leaf_count = 1;
	}

	while (true) {
		pqnode *n = m_queue.pop();
		if (n->pertinent_leaf_count == order.size())
			return n;
		pqnode *p = n->parent;
		if (p->visited_count == 0)
			p->pertinent_leaf_count = n->pertinent_leaf_count;
		else 
			p->pertinent_leaf_count += n->pertinent_leaf_count;
		p->visited_count++;
		if (p->visited_count == p->child_count)
			m_queue.push(p);
	}
	return 0; // should never happen.
}

bool
pqtree::reduce(uintvector & order)
{
	pqnode *root = find_root(order);

	reset_counter();

	m_queue.reset();
	for (uint32_t i = 0; i < m_leaves.size(); i++) {
		m_leaves[i].label = pqlabel_empty;
		m_queue.push(&m_leaves[i]);
	}

	for (uint32_t i = 0; i < order.size(); i++) {
		m_leaves[order[i]].label = pqlabel_full;
	}

	bool isroot = false;

	while (!isroot) {
		pqnode *n = m_queue.pop();
		pqnode *p = n->parent;
		isroot = n == root;

		if (!process(n, isroot)) {
			return false; // order is incompatible with the current pq-tree, should never happen!
		}

		//print();
		//printf("\n");

		p->visited_count++;
		if (p->visited_count == p->child_count)
			m_queue.push(p);
	}
	return true;
}

bool
pqtree::process(pqnode *n, bool isroot)
{
	if (n->type == pqtype_leaf)
		return true;

	uint32_t empty_count = 0;
	uint32_t full_count = 0;
	uint32_t partial_count = 0;
	pqnode *partials[2];

	for (pqnode *c = n->first; c; c = c->right) {
		switch (c->label) {
			case pqlabel_empty: 
				empty_count++;
				break;
			case pqlabel_full: 
				full_count++;
				break;
			case pqlabel_partial: 
				if (partial_count > 1) return false;
				partials[partial_count] = c;
				partial_count++;
				break;
		}
	}

	if (empty_count == 0 && partial_count == 0)
		n->label = pqlabel_full;
	else if (full_count == 0 && partial_count == 0)
		n->label = pqlabel_empty;
	else if (n->type == pqtype_p) {
		if (isroot) {
			if (partial_count == 0)
				template_p2(n);
			else if (partial_count == 1)
				template_p4(n, empty_count, full_count, partials[0]);
			else if (partial_count == 2)
				template_p6(n, empty_count, full_count, partials[0], partials[1]);
		}
		else {
			if (partial_count == 0)
				template_p3(n, empty_count, full_count);
			else if (partial_count == 1)
				template_p5(n, empty_count, full_count, partials[0]);
			else
				return false;
		}
	}
	else if (n->type == pqtype_q) {
		if (partial_count == 0)
			template_q1(n);
		else if (partial_count == 1)
			template_q2(n, partials[0]);
		else if (partial_count == 2 && isroot)
			template_q3(n, partials[0], partials[1]);
		else
			return false;
	}


	return true;
}


// n = root p node, all children are either empty or full, there is at least one of both
void
pqtree::template_p2(pqnode *n)
{
	//printf("p2\n");
	pqnode *target = new_node(pqtype_p);
	target->label = pqlabel_partial; // not really needed
	n->migrate_empty(target);
	target->insert_after(n->parent, n);
	n->detach();
	n->insert_after(target, target->last);
}


// n = root p node, all children are either empty or full, there is at least one of both, one partial
void
pqtree::template_p4(pqnode *n, uint32_t empty_count, uint32_t full_count, pqnode *partial)
{
	//printf("p4\n");
	pqnode *target = new_node(pqtype_p);
	n->label = pqlabel_partial; // not really needed
	if (full_count >= 2) {
		pqnode *target = new_node(pqtype_p);
		target->label = pqlabel_full;
		n->migrate_full(target);
		target->insert_after(partial, partial->last);
	}
	else {
		n->migrate_full(partial);
	}

	if (empty_count == 0) {
		// n has only one child, partial. replace n with partial
		partial->detach();
		partial->insert_after(n->parent, n);
		n->detach();
		release_node(n);
	}

}

// n = root p node, all children are either empty or full, there is at least one of both, two partials
void
pqtree::template_p6(pqnode *n, uint32_t empty_count, uint32_t full_count, pqnode *p1, pqnode *p2)
{
	//printf("p6\n");
	n->label = pqlabel_partial; // not really needed
	if (full_count >= 2) {
		pqnode *target = new_node(pqtype_p);
		target->label = pqlabel_full;
		n->migrate_full(target);
		target->insert_after(p1, p1->last);
	}
	else
		n->migrate_full(p1);

	p2->migrate_mirror(p1); // reverses also the order
	p2->detach();
	release_node(p2);

	if (empty_count == 0) {
		// n has only one child, p1. replace n with p1
		p1->detach();
		p1->insert_after(n->parent, n);
		n->detach();
		release_node(n);
	}

}

// n = non-root p node, all children are either empty or full, there is at least one of both
void
pqtree::template_p3(pqnode *n, uint32_t empty_count, uint32_t full_count)
{
	//printf("p3\n");
	n->type = pqtype_q;
	n->label = pqlabel_partial;

	if (full_count > 1) {
		pqnode *target = new_node(pqtype_p);
		target->label = pqlabel_full;
		n->migrate_full(target);
		target->insert_after(n, 0);
	}
	if (empty_count > 1) {
		pqnode *target = new_node(pqtype_p);
		target->label = pqlabel_empty;
		n->migrate_empty(target);
		target->insert_after(n, 0);
	}

	// make sure that empty node comes before full node
	pqnode *f = n->first;
	if (f->label == pqlabel_full) {
		f->detach();
		f->insert_after(n, n->last);
	}
}

// n = non-root p node, all children are either empty or full, there is at least one of both, one partial
void
pqtree::template_p5(pqnode *n, uint32_t empty_count, uint32_t full_count, pqnode *partial)
{
	//printf("p5\n");
	// n will be deleted by the end of this function
	if (full_count >= 2) {
		pqnode *target = new_node(pqtype_p);
		target->label = pqlabel_full;
		n->migrate_full(target);
		target->insert_after(partial, partial->last);
	}
	else
		n->migrate_full(partial);

	if (empty_count >= 2) {
		pqnode *target = new_node(pqtype_p);
		target->label = pqlabel_empty;
		n->migrate_empty(target);
		target->insert_after(partial, 0);
	}
	else
		n->migrate_empty(partial);

	// n has only one child, partial. replace n with partial
	partial->detach();
	partial->insert_after(n->parent, n);
	n->detach();
	release_node(n);
}


// n = q node, all children are either empty or full, there is at least one of both, no partials
void
pqtree::template_q1(pqnode *n)
{
	//printf("q1\n");
	if (n->first->label == pqlabel_full) {
		// need to reverse the order of the childen
		n->reverse();
	}
	n->label = pqlabel_partial;
}

// n = q node, all children are either empty or full, there is at least one of both, one partial
void
pqtree::template_q2(pqnode *n, pqnode *partial)
{
	//printf("q2\n");
	if (partial->left && partial->left->label == pqlabel_full) {
		// reverse the order so that the empty children are first
		n->reverse();
	}

	partial->migrate_after(n, partial->left);
	partial->detach();
	release_node(partial);

	/*
	n->migrate_empty(partial);
	n->migrate_full(partial);
	partial->detach();
	partial->insert_after(n->parent, n);
	n->detach();
	release_node(n);*/
}

// n = root q node, all children are either empty or full, there is at least one of both, two partials
void
pqtree::template_q3(pqnode *n, pqnode *p1, pqnode *p2)
{
	//printf("q3\n");
	n->label = pqlabel_partial; // not really needed
	// move full nodes and p2 to p1
	n->migrate_full(p1);
	p2->migrate_mirror(p1);
	p2->detach();
	release_node(p2);

	// move children of p1 to n, before p1
	p1->migrate_after(n, p1->left);
	p1->detach();
	release_node(p1);
}


pqnode *
pqtree::new_node(pqtype t)
{
	pqnode *n = m_freenode;
	n->type = t;
	n->parent = 0;

	m_freenode = n->next_free;

	return n;
}

void
pqtree::release_node(pqnode *n)
{
	n->next_free = m_freenode;
	m_freenode = n;
}
