#include "pqtree.h"
#include "dataset.h"
#include <getopt.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

/* Global variable for verbosity level */
int verbosity = 0;

void
verb_print(const int minlevel, const char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  if (verbosity >= minlevel) vfprintf(stderr, format, ap);
  va_end(ap);
}

uintvector
find_ones(const boolvector & v)
{
	uint32_t cnt = 0;
	for (uint32_t i = 0; i < v.size(); i++)
		cnt += v[i];

	uintvector res(cnt);

	for (uint32_t i = 0, j = 0; i < v.size(); i++) {
		//printf("%d ", int(v[i]));
		if (v[i]) {
			res[j] = i;
			j++;
		}
	}

	//printf("\n");

	return res;
}

uintvector
find_zeros(const boolvector & v)
{
	uint32_t cnt = 0;
	for (uint32_t i = 0; i < v.size(); i++)
		cnt += 1 - v[i];

	uintvector res(cnt);

	for (uint32_t i = 0, j = 0; i < v.size(); i++) {
		if (!v[i]) {
			res[j] = i;
			j++;
		}
	}

	return res;
}

struct block {
	uintvector row;
	uintvector col;
};

struct tile
{
	double gain;
	block raw;
	block norm;
};


class tiler
{
	public:
		tiler(dataset & d, dataset & seed, bool cyclic) :
			m_data(d), 
			m_cover(d.rowcnt(), d.colcnt()),
			m_tree(d.colcnt()),
			m_cyclic(cyclic),
			m_seed(seed)
			{}

		void iterate(const uintvector & seed, tile & t) const;
		void iterate_cyclic(const uintvector & seed, tile & t) const;
		void step_exhaustive(tile & t);
		void find_cover(uint32_t k);

		void save_tiles(const char *name);
		void save_order(const char *name);

	protected:
		double find_optimal_row(pqtree & tree, const uintvector & r, boolvector & c) const;
		double find_optimal_reverse_row(pqtree & tree, const uintvector & r, boolvector & c) const;

		//bool test_seed(const uintvector & seed);

		dataset & m_data;
		dataset m_cover;
		pqtree m_tree;

		bool m_cyclic;

		dataset & m_seed;

		typedef std::list<block> blocklist;

		blocklist m_blocks;
};


void
tiler::save_tiles(const char *name)
{
	FILE *f = fopen(name, "w");

	for (blocklist::iterator it = m_blocks.begin(); it != m_blocks.end(); ++it) {
		for (uint32_t i = 0; i < it->row.size(); i++)
			fprintf(f, "%d ", it->row[i]);
		fprintf(f, "\n");
		for (uint32_t i = 0; i < it->col.size(); i++)
			fprintf(f, "%d ", it->col[i]);
		fprintf(f, "\n");
	}

	fclose(f);
}

void
tiler::save_order(const char *name)
{
	FILE *f = fopen(name, "w");
	m_tree.print_order(f);
	fprintf(f, "\n");
	fclose(f);
}




double
tiler::find_optimal_row(pqtree & tree, const uintvector & r, boolvector & c) const
{
	doublevector w(m_data.colcnt(), -double(r.size()));
	m_data.sum_cols(r, 2, w);
	m_cover.sum_cols(r, 1, w);

	tree.set_weights(w);
	double gain = tree.find_max();
	tree.assign_optimal(c);

	return gain;
}


double
tiler::find_optimal_reverse_row(pqtree & tree, const uintvector & r, boolvector & c) const
{
	doublevector w(m_data.colcnt(), double(r.size()));
	m_data.sum_cols(r, -2, w);
	m_cover.sum_cols(r, -1, w);

	tree.set_weights(w);
	double gain = tree.find_max();
	tree.assign_optimal(c);

	double tw = 0;
	for (uint32_t i = 0; i < w.size(); i++)
		tw += w[i];

	return gain - tw;
}


void
tiler::iterate(const uintvector & seed, tile & t) const
{
	double last_gain = -1;
	t.gain = 0;

	t.raw.row = seed;

	while (t.gain > last_gain) {
		last_gain = t.gain;
		t.raw.col.swap(t.raw.row);

		pqtree clone(m_tree);
		clone.reduce(t.raw.col);

		boolvector br(m_data.rowcnt());
		t.gain = find_optimal_row(clone, t.raw.col, br);
		t.raw.row = find_ones(br);

		//for (uint32_t i = 0; i < t.raw.col.size(); i++)
			//printf("%d ", t.raw.col[i]);
		//printf("\n");
		//for (uint32_t i = 0; i < t.raw.row.size(); i++)
			//printf("%d ", t.raw.row[i]);
		//printf("\n");
	}

	t.norm = t.raw;
}

void
tiler::iterate_cyclic(const uintvector & seed, tile & t) const
{
	double last_gain = -1;
	t.gain = 0;

	t.raw.row = seed;
	t.norm.row = seed;

	double g1, g2;
	while (t.gain > last_gain) {
		last_gain = t.gain;
		t.raw.col = t.raw.row;
		t.norm.col = t.norm.row;

		pqtree clone(m_tree);
		clone.reduce(t.raw.col);

		boolvector br1(m_data.rowcnt());
		boolvector br2(m_data.rowcnt());
		g1 = find_optimal_row(clone, t.norm.col, br1);
		g2 = find_optimal_reverse_row(clone, t.norm.col, br2);
		//printf("row: %f %f\n", g1, g2);

		if (g1 > g2) {
			t.norm.row = t.raw.row = find_ones(br1);
		}
		else {
			t.raw.row = find_ones(br2);
			t.norm.row = find_zeros(br2);
		}

		//for (uint32_t i = 0; i < t.norm.row.size(); i++)
			//printf("%d ", t.norm.row[i]);
		//printf("\n");

		t.gain = std::max(g1, g2);
	}
}


void
tiler::step_exhaustive(tile & best)
{
  best.gain = -1;
#pragma omp parallel
  {
    tile priv_best;
    priv_best.gain = -1;
#pragma omp for nowait
    for (uint32_t i = 0; i < m_seed.rowcnt(); i++) {
	  const uintset & rs = m_seed.row(i);
	  uintvector seed(rs.size());
	  std::copy(rs.begin(), rs.end(), seed.begin());

      //printf("SEED: %d\n", i);
      tile t;
      
      if (i % 1000 == 0) verb_print(4, ".");
      
      if (m_cyclic)
        iterate_cyclic(seed, t);
      else
        iterate(seed, t);
      
      if (t.gain >= priv_best.gain)
        priv_best = t;
    }
#pragma omp critical
    {
      if (priv_best.gain >= best.gain)
        best = priv_best;
    }
  }
}


void
tiler::find_cover(uint32_t k)
{
	for (uint32_t i = 0; i < k; i++) {
		tile t;
    
    verb_print(2, "\tFinding tile %u... ", i);
    
    verb_print(3, "exhaustive... ");
		step_exhaustive(t);

    verb_print(3, "reducing...");
		m_tree.reduce(t.raw.row);
		m_tree.reduce(t.raw.col);
		//m_tree.print();

    verb_print(3, " removing tile... ");
		m_data.remove_tile(t.norm.row, t.norm.col);
		m_data.remove_tile(t.norm.col, t.norm.row);

    verb_print(3, "adding tile... ");
		m_cover.add_tile(t.norm.row, t.norm.col);
		m_cover.add_tile(t.norm.col, t.norm.row);

		m_blocks.push_back(t.norm);
    verb_print(2, "\n");
	}
}



int
main(int argc, char **argv)
{
	static struct option longopts[] = {
		{"out",             required_argument,  NULL, 'o'},
		{"in",              required_argument,  NULL, 'i'},
		{"comps",           required_argument,  NULL, 'k'},
		{"order",           required_argument,  NULL, 'r'},
		{"seed",            required_argument,  NULL, 's'},
		{"cyclic",           no_argument,       NULL, 'c'},
    {"verbose",         optional_argument,  NULL, 'v'},
		{ NULL,             0,                  NULL,  0 }
	};

	char *inname = NULL;
	char *outname = NULL;
	char *ordername = NULL;
	bool cyclic = false;
	uint32_t groupcnt = 1;
	dataset *seed = 0;


	int ch;
	while ((ch = getopt_long(argc, argv, "chk:o:i:v:r:s:", longopts, NULL)) != -1) {
		switch (ch) {
			case 'h':
				printf("Usage: %s -i <input file> [-o <output file>] -k <number> [-h]\n", argv[0]);
				printf("  -h    print this help\n");
        printf("  -v    increase verbosity (more -v's is more verbose)\n");
				printf("  -i    input file\n");
				printf("  -o    output file\n");
				printf("  -c    allow cyclic order\n");
				printf("  -k    number of groups\n");
				return 0;
				break;
			case 'i':
				inname = optarg;
				break;
			case 'c':
				cyclic = true;
				break;
			case 'o':
				outname = optarg;
				break;
			case 'r':
				ordername = optarg;
				break;
			case 's':
				seed = new dataset(optarg);
				break;
			case 'k':
				groupcnt = atoi(optarg);
				break;
            case 'v':
                if (optarg == NULL) verbosity++;
                else verbosity = atoi(optarg);
                break;
		}

	}

	/*
	pqtree pq(6);
	pq.print();
	uintvector r(3);
	r[0] = 3;
	r[1] = 4;
	r[2] = 5;
	pq.reduce(r);
	pq.print();

	return 0;
	*/

	if (inname == NULL) {
		printf("Missing input file\n");
		return 1;
	}
    //printf("Verbosity level is %i\n", verbosity);
	verb_print(1, "Initializing\n");
	verb_print(2, "\tLoading data from %s... ", inname);
	dataset data(inname);

	verb_print(2, "done\n");
	verb_print(2, "\tInitializing tiler... ");

	if (seed == 0) {
		seed = new dataset(data.colcnt(), data.colcnt());
		seed->add_diagonal();
	}

	tiler t(data, *seed, cyclic);

	verb_print(2, "done\n");

	verb_print(1, "Finding cover\n");
	t.find_cover(groupcnt);

	if (outname)
		t.save_tiles(outname);
	if (ordername)
		t.save_order(ordername);

	return 0;
}
